/*
 * Disk.h
 *
 *  Created on: Oct 18, 2009
 *      Author: chwolfe2
 */

#ifndef DISK_H_
#define DISK_H_
#include <stdlib.h>
#include "Defines.h"

typedef int DriveType;
#define DRIVETYPE_UNKNOWN					0
#define DRIVETYPE_SYSTEM_BLOCK_DEVICE		1
#define DRIVETYPE_SYSTEM_PARTITION			2
#define DRIVETYPE_SYSTEM_OPTICAL			3
#define DRIVETYPE_SYSTEM_FLOPPY				4
#define DRIVETYPE_INTERNAL_BLOCK_DEVICE		5
#define DRIVETYPE_INTERNAL_MULTIDISK_DEVICE	6
#define DRIVETYPE_EXTERNAL_BLOCK_DEVICE		7
#define DRIVETYPE_EXTERNAL_PARTITION		8

#define MAX_NUM_DEVICES_TO_ENUMERATE 50
#define MAX_DEVICE_NAME_LENGTH				64
#define EXT_FILE_MAX_NAME_LENGTH			128

typedef struct __Disk{
	char  		deviceName[MAX_DEVICE_NAME_LENGTH];
	boolean		locked;
	DriveType	type;
	size_t 		bytesUsed;
	size_t 		bytesTotal;
	boolean		mounted;
	dev_t		devNumber;
	//char		mountpoint[MAX_DEVICE_NAME_LENGTH];
	//size_t		numberOfChildren;
	//char  		parent[MAX_DEVICE_NAME_LENGTH];
} Disk;
#define DiskSize sizeof(Disk)

StatusCode Disk_IdentifyAll();
StatusCode Disk_GetDiskInfo(char *  deviceName, Disk* disk);
StatusCode Disk_CreateArray(Disk disksToUse[], int count);
StatusCode Disk_Format(char* deviceName);
StatusCode Disk_MountByName(char* deviceName);
StatusCode Disk_Mount(Disk* disk);
StatusCode Disk_UnmountByName(char* deviceName);
StatusCode Disk_Unmount(Disk* disk);

int DiskGetType(int index);
size_t DiskGetFreeSpace(int index);
int DiskFindIndexByName(char * deviceName);
size_t Disk_GetArrayNumDisks(Disk* disk);
size_t Disk_GetArraySize(Disk* disk);
size_t Disk_GetArrayChunkSize(Disk* disk);

extern const char *DriveTypeString[];

int    DiskGetUsableCount();
char * DiskGetUsableName(int index);
size_t DiskGetUsableFreeSpace(int index);
size_t DiskGetUsableTotalSpace(int index);

#endif /* DISK_H_ */
