/*
 * Main.c
 *
 *  Created on: Oct 17, 2009
 *      Author: chwolfe2
 */
//#include <gdbm.h>
#include <assert.h>
#include <stdio.h>
#include <string.h>
//#include <sys/mman.h>
#include "Message.h"
#include "Commands.h"
//#include "Defines.h"
//#include "Disk.h"
#include "Time.h"
#include "Log.h"
//#include "FileSystem.h"
//#include "Socket.h"
//#include "RingQueue.h"
//#include "MessageQueue.h"
#include "Persistence.h"
#include "Schedule.h"
#include "Config.h"
#include "DataFormats.h"
#include "HostInterface.h"

#define __IN_MAIN__
#include "Globals.h"
#undef  __IN_MAIN__


void Process_Messaging();

/*

StatusCode Disk_IdentifyAll();
StatusCode Disk_CreateArray(Disk disksToUse[], int count);
StatusCode Disk_Format(char* deviceName);
StatusCode Disk_Mount(char* deviceName);
StatusCode Disk_Unmount(char* deviceName);

StatusCode MessageQueue_Create(MessageQueue* msgQ, char* name);
StatusCode MessageQueue_Destroy(MessageQueue* msgQ);
StatusCode MessageQueue_Enqueue(MessageQueue* msgQ, Message* msg);
StatusCode MessageQueue_Dequeue(MessageQueue* msgQ, Message* msg);

StatusCode RingQueue_Create(RingQueue* rq, size_t insertionSize);
StatusCode RingQueue_Destroy(RingQueue* rq);
StatusCode RingQueue_NextInsertionPoint(RingQueue* rq, char* insertionPoint);
StatusCode RingQueue_NotifyInsertionComplete(RingQueue* rq, size_t bytesInserted);
StatusCode RingQueue_NextRemovalPoint(RingQueue* rq, char* removalPoint);
StatusCode RingQueue_NotifyRemovalComplete(RingQueue* rq, size_t bytesRemoved);

StatusCode Socket_Create(Socket* s);
StatusCode Socket_OpenRecieve(Socket* s, unsigned short port);
StatusCode Socket_OpenSend(Socket* s, char* destinationURL, unsigned short port);
StatusCode Socket_Recieve(Socket* s, char* buffer, size_t* size);
StatusCode Socket_Send(Socket* s, char* buffer, size_t size);
StatusCode Socket_Close(Socket* s);
StatusCode Socket_Destroy(Socket* s);

StatusCode DataFormats_Initialize		(boolean reloadConfig);
StatusCode DataFormats_GetCount			(int* count);
StatusCode DataFormats_GetName			(int index, char * formatName);
StatusCode DataFormats_GetSpec			(int index, char * spec);
StatusCode DataFormats_GetPayloadSize	(int index, size_t * payloadSize);
StatusCode DataFormats_GetRate			(int index, size_t * rate);
StatusCode DataFormats_Get				(char * formatName, DataFormat * result);
StatusCode DataFormats_Compile			(DataFormat * format, CompiledDataFormat* result);
StatusCode DataFormats_Release			(CompiledDataFormat* compiledFormat);
StatusCode DataFormats_Finalize();
void printCompiledFormat(CompiledDataFormat* compiledFormat);

*/

//GDBM_FILE* database, char * fileName, char * defaultsFileName
boolean BootOption_DeleteLog=false;			// -flushLog
boolean BootOption_DeleteSchedule=false;	// -flushSchedule
boolean BootOption_DeleteData=false;		// -flushData
boolean BootOption_ResetConfigData=false;	// -flushConfig
boolean ShutdownOption_Scram = false;
boolean ShutdownOption_Reboot = true;

void ParseBootOptions(int argc, char* argv[]){
	int i = 1;
	while (i<argc){
		Log_Add("[BOOT] Processing option '%s'",argv[i]);
		if 			(strcmp(argv[i],"-flushLog")==0) 		BootOption_DeleteLog = true;
		else if 	(strcmp(argv[i],"-flushSchedule")==0) 	BootOption_DeleteSchedule = true;
		else if 	(strcmp(argv[i],"-flushData")==0) 		BootOption_DeleteData = true;
		else if 	(strcmp(argv[i],"-flushConfig")==0) 	BootOption_ResetConfigData = true;
		i++;
	}
}

int main(int argc, char* argv[]){

/* code to test process hosting
	const int sz=200;
	ProcessHandle* ph;
	StatusCode sc;
	char* comand = "ls -l /home/chwolfe2";
	char ls_output[sz];
	s =  LaunchProcess(comand, 1, sz, ls_output, &ph);
	if (s!=SUCCESS){
		printf("Can't launch process '%s'\n",comand);
	}
	while ((s= CheckProcessDone(ph))==NOT_READY){
		//wait...
	}
	if (s!=SUCCESS){
		printf("Process exited abnormally.\n");
	} else {
		printf("Process output for '%s'\n'%s'\n",comand,ls_output);
	}
	ReleaseProcessHandle(ph);
	exit(0);
//*/


	//////////////////////////////////////////////////////////////////////////////////
	//////////////////////////////////////////////////////////////////////////////////
	//
	//        @@@@@    @@@@   @@@@  @@@@@@
	//        @@  @@  @@  @@ @@  @@   @@
	//        @@@@@   @@  @@ @@  @@   @@
	//        @@  @@  @@  @@ @@  @@   @@
	//        @@@@@    @@@@   @@@@    @@
	//
	//////////////////////////////////////////////////////////////////////////////////
	//////////////////////////////////////////////////////////////////////////////////

	StatusCode sc;
	Disk 			disk;
	FileSystem 		fileSystem;
	MessageQueue 	messageQueue;
	MessageQueue 	responseQueue;
	RingQueue 		rxQueue;
	Socket 			rxMsgSocket;
	Socket 			txMsgSocket;
	Socket 			rxSocket;
	bzero((void *)&disk,			sizeof(Disk));
	bzero((void *)&fileSystem,		sizeof(FileSystem));
	bzero((void *)&messageQueue,	sizeof(MessageQueue));
	bzero((void *)&responseQueue,	sizeof(MessageQueue));
	bzero((void *)&rxQueue,			sizeof(RingQueue));
	bzero((void *)&rxMsgSocket,		sizeof(Socket));
	bzero((void *)&txMsgSocket,		sizeof(Socket));
	bzero((void *)&rxSocket,		sizeof(Socket));

	bzero((void*)&globals,sizeof(Globals));
	globals.disk			= &disk;
	globals.fileSystem		= &fileSystem;
	globals.messageQueue	= &messageQueue;
	globals.responseQueue	= &responseQueue;
	globals.rxQueue			= &rxQueue;
	globals.rxMsgSocket		= &rxMsgSocket;
	globals.txMsgSocket		= &txMsgSocket;
	globals.rxSocket		= &rxSocket;
	globals.flushConfigRequested 	= false;
	globals.flushDataRequested 		= false;
	globals.flushLogRequested 		= false;
	globals.flushScheduleRequested 	= false;
	globals.exitRequested			= false;
	globals.shutdownRequested 		= false;
	globals.rebootRequested 		= true;
	globals.scramRequested 			= false;


	DataFormat 			df;
	CompiledDataFormat 	cdf;
	Disk 				extdisk;
	MyFile				tagfile;
	bzero((void *)&df,			sizeof(DataFormat));
	bzero((void *)&cdf,			sizeof(CompiledDataFormat));
	bzero((void *)&extdisk,		sizeof(Disk));
	bzero((void *)&tagfile,		sizeof(MyFile));
	globals.opData.compiledFormat 		= &cdf;
	globals.opData.dataFormat			= &df;
	globals.opData.extdisk				= &extdisk;
	globals.opData.tagfile				= &tagfile;
	globals.opData.initialParameters	= NULL;
	globals.opData.errorCount 			= 0l;
	globals.opData.warningCount 		= 0l;


	globals.LocalEchoMessaging			= true;


	//parse command line options
	ParseBootOptions(argc,argv);
	Log_Add("[BOOT] Loading system configuration.");






	// load configuration
	sc=Config_Initialize(BootOption_ResetConfigData);
	if (sc!=SUCCESS){
		Log_Add("[BOOT] Fatal Error: Could not load system configuration. Exiting...");
		return EXIT_CODE_FAIL;
	}
	char temp[8192];
	Config_Get("MessageInPort",temp);
	globals.MessageInPort = (unsigned short) strtoul(temp,NULL,10);
	Config_Get("MessageOutPort",temp);
	globals.MessageOutPort = (unsigned short) strtoul(temp,NULL,10);
	Config_Get("DataInPort",temp);
	globals.DataInPort = (unsigned short) strtoul(temp,NULL,10);
	Config_Get("MessageOutURL",globals.MessageOutURL);
	Config_Get("MyReferenceDesignator", globals.MyReferenceDesignator);
	Config_Get("MySerialNumber", globals.MySerialNumber);
	Config_Get("Version", globals.Version);
	Config_Get("TimeAuthority",globals.ntpServer);
	Config_Get("ArraySelect",temp);
	globals.arraySelect = (unsigned short) strtoul(temp,NULL,10);




	char command[8192];
	char result[8192];
	//TODO: remove this and configure ntpd properly
	issueShellCommand("killall ntpd", result, 8192);
	sprintf(command, "ntpdate %s", globals.ntpServer);
	sc= ShellGetString(command, result, 8192);
	if (sc!=SUCCESS){
		Log_Add("[BOOT] Error: NTP synchronization process failed");
	} else {
		Log_Add("[BOOT] NTP synchronization result: '%s'",result);
	}


	// initialize the log
	Log_Add("[BOOT] Initializing the system log.");
	sc = Log_Initialize(BootOption_DeleteLog);
	if (sc!=SUCCESS){
		Log_Add("[BOOT] Fatal Error: Could not initialize system log. Exiting...");
		Config_Finalize();
		return EXIT_CODE_FAIL;
	}
	Log_Add("[BOOT] Opened System Log.");


	// initialize the drive devices
	Log_Add("[BOOT] Initializing the system's storage devices.");
	sc = Disk_IdentifyAll();
	if (sc!=SUCCESS){
		Log_Add("[BOOT] Fatal Error: Could not enumerate system storage devices.Exiting...");
		Log_Close();
		Config_Finalize();
		return EXIT_CODE_FAIL;
	}
	Log_Add("[BOOT] Initialized system's storage devices.");

	// find internal storage looking for /dev/md0, then /dev/md1, upto /dev/md5
	Log_Add("[BOOT] Locating internal storage among detected devices.");
	int arrayNum=0;
	boolean arrayFound=false;
	char arrayDeviceName[128];

	char arrayNames[10][128];
	unsigned short numArraysFound=0;
	Disk arrays[10];
	while (arrayNum < 6){
		sprintf(arrayDeviceName, "/dev/md%d", arrayNum);
		sc = Disk_GetDiskInfo(arrayDeviceName, &arrays[numArraysFound]);
		if (sc==SUCCESS){
			arrayFound=true;
			strncpy(arrayNames[numArraysFound],arrayDeviceName,128);
			numArraysFound++;
		}
		arrayNum++;
	}
	Log_Add("[BOOT] Information: Located %hu internal storage devices (DRSUs).",numArraysFound);

	if (!arrayFound){
		Log_Add("[BOOT] Fatal Error: Could not locate internal storage among detected devices. Exiting...");
		Log_Close();
		Config_Finalize();
		return EXIT_CODE_FAIL;
	} else {
		globals.numArraysFound=numArraysFound;
		if (globals.arraySelect>globals.numArraysFound){
			Log_Add("[BOOT] Error: boot config requested DRSU %hu, but only found %hu DRSUs. Defaulting to DRSU 0.",globals.arraySelect,globals.numArraysFound);
			globals.arraySelect=0;
		}
		globals.arrays = arrays;
		memcpy(&disk,&arrays[globals.arraySelect],sizeof(Disk));
		Log_Add("[BOOT] Identified internal storage device as '%s'",disk.deviceName);
	}

	// possibly create filesystem
	if (BootOption_DeleteData){
		Log_Add("[BOOT] Creating internal storage file system.");
		sc = FileSystem_Create(&disk);
		if (sc!=SUCCESS){
			Log_Add("[BOOT] Fatal Error: Could not create internal storage file system. Exiting...");
			Log_Close();
			Config_Finalize();
			return EXIT_CODE_FAIL;
		} else {
			Log_Add("[BOOT] File system created, all data deleted!!!");
		}
	}

	// open filesystem
	Log_Add("[BOOT] Opening internal storage file system.");
	sc = FileSystem_Open(&fileSystem, &disk);
	if (sc!=SUCCESS){
		Log_Add("[BOOT] Fatal Error: Could not open internal storage file system. Exiting...");
		Log_Close();
		Config_Finalize();
		return EXIT_CODE_FAIL;
	}
	Log_Add("[BOOT] Opened internal storage file system.");
	/*printf("globals.fileSystem = %lu \n",(size_t) globals.fileSystem);
	printf("globals.fileSystem->fileSystemHeaderData = %lu \n",(size_t) globals.fileSystem->fileSystemHeaderData);
	printf("globals.fileSystem->fileSystemHeaderData->filesystemInfo.chunkSize = %lu \n",(size_t) globals.fileSystem->fileSystemHeaderData->filesystemInfo.chunkSize);
*/
	// initialize data formats
	Log_Add("[BOOT] Opening data formats database.");
	sc = DataFormats_Initialize(BootOption_ResetConfigData);
	if (sc!=SUCCESS){
		Log_Add("[BOOT] Fatal Error: Could not read data formats database. Exiting...");
		FileSystem_Close(&fileSystem);
		Log_Close();
		Config_Finalize();
		return EXIT_CODE_FAIL;
	}
	Log_Add("[BOOT] Read data formats database.");


	// create messaging receive socket
	Log_Add("[BOOT] Creating messaging receive socket.");
	sc = Socket_Create(&rxMsgSocket);
	if (sc!=SUCCESS){
		Log_Add("[BOOT] Fatal Error: Could not create messaging receive socket. Exiting...");
		DataFormats_Finalize();
		FileSystem_Close(&fileSystem);
		Log_Close();
		Config_Finalize();
		return EXIT_CODE_FAIL;
	}
	Log_Add("[BOOT] Created messaging receive socket.");

	// create messaging send socket
	Log_Add("[BOOT] Creating messaging send socket.");
	sc = Socket_Create(&txMsgSocket);
	if (sc!=SUCCESS){
		Log_Add("[BOOT] Fatal Error: Could not create messaging send socket. Exiting...");
		Socket_Destroy(&rxMsgSocket);
		DataFormats_Finalize();
		FileSystem_Close(&fileSystem);
		Log_Close();
		Config_Finalize();
		return EXIT_CODE_FAIL;
	}
	Log_Add("[BOOT] Created messaging send socket.");

	// open messaging receive socket
	Log_Add("[BOOT] Opening messaging receive socket.");
	sc = Socket_OpenRecieve(&rxMsgSocket,globals.MessageInPort);
	if (sc!=SUCCESS){
		Log_Add("[BOOT] Fatal Error: Could not open messaging receive socket. Exiting...");
		Socket_Destroy(&txMsgSocket);
		Socket_Destroy(&rxMsgSocket);
		DataFormats_Finalize();
		FileSystem_Close(&fileSystem);
		Log_Close();
		Config_Finalize();
		return EXIT_CODE_FAIL;
	}
	Log_Add("[BOOT] Opened messaging receive socket.");

	// open messaging send socket
	Log_Add("[BOOT] Opening messaging send socket.");
	sc = Socket_OpenSend(&txMsgSocket,globals.MessageOutURL,globals.MessageOutPort);
	if (sc!=SUCCESS){
		Log_Add("[BOOT] Fatal Error: Could not open messaging send socket. Exiting...");fflush(stdout);
		Socket_Close(&rxMsgSocket);
		Socket_Destroy(&txMsgSocket);
		Socket_Destroy(&rxMsgSocket);
		DataFormats_Finalize();
		FileSystem_Close(&fileSystem);
		Log_Close();
		Config_Finalize();
		return EXIT_CODE_FAIL;
	}
	Log_Add("[BOOT] Opened messaging send socket.");fflush(stdout);

	// create messaging receive queue
	Log_Add("[BOOT] Creating messaging receive queue.");fflush(stdout);
	sc = MessageQueue_Create(&messageQueue,"/message");
	if (sc!=SUCCESS){
		Log_Add("[BOOT] Fatal Error: Could not create messaging receive queue. Exiting...");
		Socket_Close(&txMsgSocket);
		Socket_Close(&rxMsgSocket);
		Socket_Destroy(&txMsgSocket);
		Socket_Destroy(&rxMsgSocket);
		DataFormats_Finalize();
		FileSystem_Close(&fileSystem);
		Log_Close();
		Config_Finalize();
		return EXIT_CODE_FAIL;
	}
	Log_Add("[BOOT] Created messaging receive queue.");

	// create messaging send queue
	Log_Add("[BOOT] Creating messaging send queue.");
	sc = MessageQueue_Create(&responseQueue,"/response");
	if (sc!=SUCCESS){
		Log_Add("[BOOT] Fatal Error: Could not create messaging send queue. Exiting...");
		MessageQueue_Destroy(&messageQueue);
		Socket_Close(&txMsgSocket);
		Socket_Close(&rxMsgSocket);
		Socket_Destroy(&txMsgSocket);
		Socket_Destroy(&rxMsgSocket);
		DataFormats_Finalize();
		FileSystem_Close(&fileSystem);
		Log_Close();
		Config_Finalize();
		return EXIT_CODE_FAIL;
	}
	Log_Add("[BOOT] Created messaging send queue.");


	// create data receive socket
	Log_Add("[BOOT] Creating data receive socket.");
	sc = Socket_Create(&rxSocket);
	if (sc!=SUCCESS){
		Log_Add("[BOOT] Fatal Error: Could not create data receive socket. Exiting...");
		MessageQueue_Destroy(&responseQueue);
		MessageQueue_Destroy(&messageQueue);
		Socket_Close(&txMsgSocket);
		Socket_Close(&rxMsgSocket);
		Socket_Destroy(&txMsgSocket);
		Socket_Destroy(&rxMsgSocket);
		DataFormats_Finalize();
		FileSystem_Close(&fileSystem);
		Log_Close();
		Config_Finalize();
		return EXIT_CODE_FAIL;
	}
	Log_Add("[BOOT] Created data receive socket.");

	// create data receive queue
	Log_Add("[BOOT] Creating data receive queue.");
	sc = RingQueue_Create(&rxQueue,16384);
	if (sc!=SUCCESS){
		Log_Add("[BOOT] Fatal Error: Could not create data receive queue. Exiting...");
		Socket_Destroy(&rxSocket);
		MessageQueue_Destroy(&responseQueue);
		MessageQueue_Destroy(&messageQueue);
		Socket_Close(&txMsgSocket);
		Socket_Close(&rxMsgSocket);
		Socket_Destroy(&txMsgSocket);
		Socket_Destroy(&rxMsgSocket);
		DataFormats_Finalize();
		FileSystem_Close(&fileSystem);
		Log_Close();
		Config_Finalize();
		return EXIT_CODE_FAIL;
	}
	Log_Add("[BOOT] Created data receive queue.");



	// open schedule
	Log_Add("[BOOT] Reading system schedule.");
	sc = Schedule_Open(BootOption_DeleteSchedule);
	if (sc!=SUCCESS){
		Log_Add("[BOOT] Fatal Error: Read system schedule. Exiting...");
		RingQueue_Destroy(&rxQueue);
		Socket_Destroy(&rxSocket);
		MessageQueue_Destroy(&responseQueue);
		MessageQueue_Destroy(&messageQueue);
		Socket_Close(&txMsgSocket);
		Socket_Close(&rxMsgSocket);
		Socket_Destroy(&txMsgSocket);
		Socket_Destroy(&rxMsgSocket);
		DataFormats_Finalize();
		FileSystem_Close(&fileSystem);
		Log_Close();
		Config_Finalize();
		return EXIT_CODE_FAIL;
	}

	Log_Add("[BOOT] Finished reading system schedule.");
	Log_Add("[BOOT] Booting Complete.");
	globals.upStatus=UP;



	//////////////////////////////////////////////////////////////////////////////////
	//////////////////////////////////////////////////////////////////////////////////
	//
	//        @@@@  @@@@   @@@@@   @@@@@@  @@   @@      @@     @@@@   @@@@  @@@@
	//        @@ @@@@ @@  @@   @@    @@    @@@  @@      @@    @@  @@ @@  @@ @@ @@
	//        @@  @@  @@  @@@@@@@    @@    @@ @ @@      @@    @@  @@ @@  @@ @@@@
	//        @@      @@  @@   @@    @@    @@  @@@      @@    @@  @@ @@  @@ @@
	//        @@      @@  @@   @@  @@@@@@  @@   @@      @@@@@  @@@@   @@@@  @@
	//
	//////////////////////////////////////////////////////////////////////////////////
	//////////////////////////////////////////////////////////////////////////////////

	// Main control loop
	//globals.exitRequested=true;
	while(!globals.exitRequested){
		Process_Messaging();
		Process_Schedule();
	}

	//////////////////////////////////////////////////////////////////////////////////
	//////////////////////////////////////////////////////////////////////////////////
	//
	//      @@@@@@ @@   @@ @@   @@ @@@@@@      @@@@@   @@@@  @@   @@ @@   @@
	//      @@     @@   @@ @@   @@   @@        @@  @@ @@  @@ @@   @@ @@@  @@
	//      @@@@@@ @@@@@@@ @@   @@   @@        @@  @@ @@  @@ @@ @ @@ @@ @ @@
	//          @@ @@   @@ @@@ @@@   @@        @@  @@ @@  @@ @@@@@@@ @@  @@@
	//      @@@@@@ @@   @@  @@@@@    @@        @@@@@   @@@@   @@ @@  @@   @@
	//
	//////////////////////////////////////////////////////////////////////////////////
	//////////////////////////////////////////////////////////////////////////////////

	// close schedule
	Log_Add("[SHUTDOWN] Closing system schedule.");
	sc = Schedule_Close();
	if (sc!=SUCCESS){
		Log_Add("[SHUTDOWN] Fatal Error: close system schedule. Exiting...");
		RingQueue_Destroy(&rxQueue);
		Socket_Destroy(&rxSocket);
		MessageQueue_Destroy(&responseQueue);
		MessageQueue_Destroy(&messageQueue);
		Socket_Close(&txMsgSocket);
		Socket_Close(&rxMsgSocket);
		Socket_Destroy(&txMsgSocket);
		Socket_Destroy(&rxMsgSocket);
		DataFormats_Finalize();
		FileSystem_Close(&fileSystem);
		Log_Close();
		Config_Finalize();
		return EXIT_CODE_FAIL;
	}
	Log_Add("[SHUTDOWN] Closed system schedule.");

	if (!globals.scramRequested){

		// destroy data receive queue
		Log_Add("[SHUTDOWN] Destroying data receive queue.");
		sc = RingQueue_Destroy(&rxQueue);
		if (sc!=SUCCESS){
			Log_Add("[SHUTDOWN] Fatal Error: Could not destroy data receive queue. Exiting...");
			Socket_Destroy(&rxSocket);
			MessageQueue_Destroy(&responseQueue);
			MessageQueue_Destroy(&messageQueue);
			Socket_Close(&txMsgSocket);
			Socket_Close(&rxMsgSocket);
			Socket_Destroy(&txMsgSocket);
			Socket_Destroy(&rxMsgSocket);
			DataFormats_Finalize();
			FileSystem_Close(&fileSystem);
			Log_Close();
			Config_Finalize();
			return EXIT_CODE_FAIL;
		}
		Log_Add("[SHUTDOWN] Destroyed data receive queue.");

		// destroy data receive socket
		Log_Add("[SHUTDOWN] Destroying data receive socket.");
		sc = Socket_Destroy(&rxSocket);
		if (sc!=SUCCESS){
			Log_Add("[SHUTDOWN] Fatal Error: Could not destroy data receive socket. Exiting...");
			MessageQueue_Destroy(&responseQueue);
			MessageQueue_Destroy(&messageQueue);
			Socket_Close(&txMsgSocket);
			Socket_Close(&rxMsgSocket);
			Socket_Destroy(&txMsgSocket);
			Socket_Destroy(&rxMsgSocket);
			DataFormats_Finalize();
			FileSystem_Close(&fileSystem);
			Log_Close();
			Config_Finalize();
			return EXIT_CODE_FAIL;
		}
		Log_Add("[SHUTDOWN] Destroyed data receive socket.");

		// destroy messaging send queue
		Log_Add("[SHUTDOWN] Destroying messaging send queue.");
		sc = MessageQueue_Destroy(&responseQueue);
		if (sc!=SUCCESS){
			Log_Add("[SHUTDOWN] Fatal Error: Could not destroy messaging send queue. Exiting...");
			MessageQueue_Destroy(&messageQueue);
			Socket_Close(&txMsgSocket);
			Socket_Close(&rxMsgSocket);
			Socket_Destroy(&txMsgSocket);
			Socket_Destroy(&rxMsgSocket);
			DataFormats_Finalize();
			FileSystem_Close(&fileSystem);
			Log_Close();
			Config_Finalize();
			return EXIT_CODE_FAIL;
		}
		Log_Add("[SHUTDOWN] Destroyed messaging send queue.");

		// destroy messaging receive queue
		Log_Add("[SHUTDOWN] Destroying messaging receive queue.");
		sc = MessageQueue_Destroy(&messageQueue);
		if (sc!=SUCCESS){
			Log_Add("[SHUTDOWN] Fatal Error: Could not destroy messaging receive queue. Exiting...");
			Socket_Close(&txMsgSocket);
			Socket_Close(&rxMsgSocket);
			Socket_Destroy(&txMsgSocket);
			Socket_Destroy(&rxMsgSocket);
			DataFormats_Finalize();
			FileSystem_Close(&fileSystem);
			Log_Close();
			Config_Finalize();
			return EXIT_CODE_FAIL;
		}
		Log_Add("[SHUTDOWN] Destroyed messaging receive queue.");

		// close messaging send socket
		Log_Add("[SHUTDOWN] Closing messaging send socket.");
		sc = Socket_Close(&txMsgSocket);
		if (sc!=SUCCESS){
			Log_Add("[SHUTDOWN] Fatal Error: Could not close messaging send socket. Exiting...");
			Socket_Close(&rxMsgSocket);
			Socket_Destroy(&txMsgSocket);
			Socket_Destroy(&rxMsgSocket);
			DataFormats_Finalize();
			FileSystem_Close(&fileSystem);
			Log_Close();
			Config_Finalize();
			return EXIT_CODE_FAIL;
		}
		Log_Add("[SHUTDOWN] Closed messaging send socket.");

		// close messaging receive socket
		Log_Add("[SHUTDOWN] Closing messaging receive socket.");
		sc = Socket_Close(&rxMsgSocket);
		if (sc!=SUCCESS){
			Log_Add("[SHUTDOWN] Fatal Error: Could not close messaging receive socket. Exiting...");
			Socket_Destroy(&txMsgSocket);
			Socket_Destroy(&rxMsgSocket);
			DataFormats_Finalize();
			FileSystem_Close(&fileSystem);
			Log_Close();
			Config_Finalize();
			return EXIT_CODE_FAIL;
		}
		Log_Add("[SHUTDOWN] Closed messaging receive socket.");

		// destroy messaging send socket
		Log_Add("[SHUTDOWN] Destroying messaging send socket.");
		sc = Socket_Destroy(&txMsgSocket);
		if (sc!=SUCCESS){
			Log_Add("[SHUTDOWN] Fatal Error: Could not destroy messaging send socket. Exiting...");
			Socket_Destroy(&rxMsgSocket);
			DataFormats_Finalize();
			FileSystem_Close(&fileSystem);
			Log_Close();
			Config_Finalize();
			return EXIT_CODE_FAIL;
		}
		Log_Add("[SHUTDOWN] Destroyed messaging send socket.");

		// destroy messaging send socket
		Log_Add("[SHUTDOWN] Destroying messaging receive socket.");
		sc = Socket_Destroy(&rxMsgSocket);
		if (sc!=SUCCESS){
			Log_Add("[SHUTDOWN] Fatal Error: Could not destroy messaging receive socket. Exiting...");
			DataFormats_Finalize();
			FileSystem_Close(&fileSystem);
			Log_Close();
			Config_Finalize();
			return EXIT_CODE_FAIL;
		}
		Log_Add("[SHUTDOWN] Destroyed messaging receive socket.");

		// close data formats database
		Log_Add("[SHUTDOWN] Closing data formats database.");
		sc = DataFormats_Finalize();
		if (sc!=SUCCESS){
			Log_Add("[SHUTDOWN] Fatal Error: Could not close data formats database. Exiting...");
			FileSystem_Close(&fileSystem);
			Log_Close();
			Config_Finalize();
			return EXIT_CODE_FAIL;
		}
		Log_Add("[SHUTDOWN] Closed messaging receive socket.");
	} // if "SCRAM"


	// close file system
	Log_Add("[SHUTDOWN] Closing internal storage file system.");
	sc = FileSystem_Close(&fileSystem);
	if (sc!=SUCCESS){
		Log_Add("[SHUTDOWN] Fatal Error: Could not close internal storage file system. Exiting...");
		Log_Close();
		Config_Finalize();
		return EXIT_CODE_FAIL;
	}
	Log_Add("[SHUTDOWN] Closed internal storage file system.");

	// close system log
	Log_Add("[SHUTDOWN] Closing system log.");
	Log_Close();


	// close config database
	Log_Add("[SHUTDOWN] Closing config database. ");
	sc = Config_Finalize();
	if (sc!=SUCCESS){
		Log_Add("[SHUTDOWN] Fatal Error: Could not close config database. Exiting... ");
		return EXIT_CODE_FAIL;
	}
	int exitcode = 0;

	if (globals.flushConfigRequested)exitcode |= EXIT_CODE_FLUSH_CONFIG;
	if (globals.flushDataRequested) exitcode |= EXIT_CODE_FLUSH_DATA;
	if (globals.flushLogRequested) exitcode |= EXIT_CODE_FLUSH_LOG;
	if (globals.flushScheduleRequested) exitcode |= EXIT_CODE_FLUSH_SCHEDULE;
	if (globals.shutdownRequested) exitcode |= EXIT_CODE_SHUTDOWN;
	if (globals.rebootRequested) exitcode |= EXIT_CODE_REBOOT;
	return exitcode;
}

//////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////
//
//        @@@@@  @@@@@   @@@@   @@@@@  @@@@@@ @@@@@@ @@@@@@ @@@@@@ @@@@@@
//        @@  @@ @@  @@ @@  @@ @@@  @@ @@     @@     @@     @@     @@
//        @@@@@  @@@@@  @@  @@ @@      @@@@   @@@@@@ @@@@@@ @@@@   @@@@@@
//        @@     @@  @@ @@  @@ @@@  @@ @@         @@     @@ @@         @@
//        @@     @@  @@  @@@@   @@@@@  @@@@@@ @@@@@@ @@@@@@ @@@@@@ @@@@@@
//
//////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////
static BaseMessage messageJustIn;
static BaseMessage messageReadyToProcess;
static SafeMessage messageReadyToProcessConv;
static SafeMessage responseUnConv;
static BaseMessage response;
static BaseMessage responseReadyToSend;
static boolean processed = false;
static boolean busy_messageJustIn = false;
static boolean busy_messageReadyToProcess = false;
static boolean busy_responseReadyToSend = false;
static size_t bytesReceived;
static size_t bytesToSend;

void Process_Messaging(){
	StatusCode sc;

	if (!busy_messageJustIn){
		sc=Socket_Recieve(globals.rxMsgSocket,(char*)&messageJustIn, &bytesReceived);
		if (sc==SUCCESS){
			if ( (strncmp(messageJustIn.destination,globals.MyReferenceDesignator,MESSAGE_FIELDSIZE_DESTINATION)==0) ||
				 (strncmp(messageJustIn.destination,"ALL",MESSAGE_FIELDSIZE_DESTINATION)==0)){
				busy_messageJustIn=true;
			}
		}
	}
	if (busy_messageJustIn){
		sc=MessageQueue_Enqueue(globals.messageQueue,&messageJustIn);
		if (sc==SUCCESS){
			bzero((void*)&messageJustIn,sizeof(messageJustIn));
			busy_messageJustIn=false;
		} /*else {
			printf("sc= %s\n", ((sc==FAILURE)?"FAILURE":"NOT_READY"));
		}*/
	}
	if (!busy_messageReadyToProcess){
		sc=MessageQueue_Dequeue(globals.messageQueue,&messageReadyToProcess);
		if (sc==SUCCESS){
			//((char*)&messageReadyToProcess)[8191]='\0';
			//printf("%s\n",(char*)&messageReadyToProcess);
			busy_messageReadyToProcess=true;
		}
	}

	if (busy_messageReadyToProcess){
		if (!processed){
			MessageConvertBaseSafe(&messageReadyToProcess,&messageReadyToProcessConv);
			//printf("RefSafe %s  vs.   RefBase %s",messageReadyToProcessConv.reference,messageReadyToProcess.reference);
			//if (Message_Verify(&messageReadyToProcess)){
				sc = Command_Process(&messageReadyToProcessConv,&responseUnConv);
				//printf("Command_Process(&messageReadyToProcessConv,&responseUnConv) = %s\n",statusString(sc));fflush(stdout);
				if (sc!=SUCCESS){
					Log_Add("MESSAGING: Error: could not process message.");
					//TODO: when messaging fails
				}
				processed=true;
				//printf("Processed\n");
			/*} else {
				sc = GenerateErrorResponse(&messageReadyToProcess,&response,ERROR_CODE_ILLEGAL_CHARACTERS,"");
				if (sc!=SUCCESS){
					Log_Add("MESSAGING: Error: could not process message.");
					//TODO: when messaging fails
				}
				processed=true;
			}*/
		}
		MessageConvertSafeBase(&responseUnConv,&response);
		sc=MessageQueue_Enqueue(globals.responseQueue,&response);
		if (sc==SUCCESS){
			bzero((void*)&messageReadyToProcess,sizeof(messageReadyToProcess));
			bzero((void*)&response,sizeof(response));
			busy_messageReadyToProcess=false;
			processed=false;
		}

	}
	if (!busy_responseReadyToSend){
		sc=MessageQueue_Dequeue(globals.responseQueue,&responseReadyToSend);
		if (sc==SUCCESS){
			busy_responseReadyToSend=true;
		}
	}
	if (busy_responseReadyToSend){
		bytesToSend = Message_TransmitSize(&responseReadyToSend);
		sc=Socket_Send(globals.txMsgSocket, (char *)&responseReadyToSend, bytesToSend);
		if (sc==SUCCESS){
			char buffer[256];
			snprintf(buffer,255,"%s",(char*)&responseReadyToSend);
			buffer[255]='\0';
			//Log_Add("MESSAGING: Responded with: '%s'",buffer);
			busy_responseReadyToSend=false;
		}
	}
}
void Process_Operation(){
}

void Process_Status(){
}







